//! Manage application shutdown
//!
//! This crate provides a set of types for managing graceful application shutdown.
//!
//! # Coordinate shutdown with the [`ShutdownManager`] type
//!
//! The [`ShutdownManager`] is used to coordinate shutdown of the process in an ordered fashion.
//! When a shutdown is signaled externally, e.g., `ctrl+c`, or internally by some error state, then
//! there may be processes running on the backend that need to be gracefully stopped before the
//! HTTP/gRPC frontend. For example, if the WAL has writes buffered, it needs to flush the buffer to
//! object store and respond to the write request before the HTTP/gRPC frontend is taken down.
//!
//! Components can [`register`][ShutdownManager::register] to receive a [`ShutdownToken`], which can
//! be used to [`wait_for_shutdown`][ShutdownToken::wait_for_shutdown] to trigger component-specific
//! cleanup logic before signaling back via [`complete`][ShutdownToken::complete] to indicate that
//! shutdown can proceed.

use observability_deps::tracing::info;
use parking_lot::Mutex;
use tokio::sync::oneshot;
pub use tokio_util::sync::CancellationToken;
use tokio_util::task::TaskTracker;

/// Wait for a `SIGTERM` or `SIGINT` to stop the process on UNIX systems
#[cfg(unix)]
pub async fn wait_for_signal() {
    use tokio::signal::unix::{SignalKind, signal};

    let mut term = signal(SignalKind::terminate()).expect("failed to register signal handler");
    let mut int = signal(SignalKind::interrupt()).expect("failed to register signal handler");

    tokio::select! {
        _ = term.recv() => info!("Received SIGTERM"),
        _ = int.recv() => info!("Received SIGINT"),
    }
}

/// Wait for a `ctrl+c` to stop the process on Windows systems
#[cfg(windows)]
pub async fn wait_for_signal() {
    let _ = tokio::signal::ctrl_c().await;
    info!("Received SIGINT");
}

/// Manage application shutdown
///
/// This deries `Clone`, as the underlying `tokio` types can be shared via clone.
#[derive(Debug, Clone)]
pub struct ShutdownManager {
    frontend_shutdown: CancellationToken,
    backend_shutdown: CancellationToken,
    tasks: TaskTracker,
}

impl ShutdownManager {
    /// Create a [`ShutdownManager`]
    ///
    /// Accepts a [`CancellationToken`] which the `ShutdownManager` will signal cancellation to
    /// after the backend has cleanly shutdown.
    pub fn new(frontend_shutdown: CancellationToken) -> Self {
        Self {
            frontend_shutdown,
            backend_shutdown: CancellationToken::new(),
            tasks: TaskTracker::new(),
        }
    }

    /// Create a [`ShutdownManager`] for testing purposes
    ///
    /// This handles creation of the frontend [`CancellationToken`] for tests where `tokio-util` is
    /// not a dependency, or handling of the frontend shutdown is not necessary.
    pub fn new_testing() -> Self {
        Self {
            frontend_shutdown: CancellationToken::new(),
            backend_shutdown: CancellationToken::new(),
            tasks: TaskTracker::new(),
        }
    }

    /// Register a task that needs to perform work before the process may exit
    ///
    /// Provides a [`ShutdownToken`] which the caller is responsible for handling. The caller must
    /// invoke [`complete`][ShutdownToken::complete] in order for process shutdown to succeed.
    pub fn register(&self) -> ShutdownToken {
        let (tx, rx) = oneshot::channel();
        self.tasks.spawn(rx);
        ShutdownToken::new(self.backend_shutdown.clone(), tx)
    }

    /// Waits for registered tasks to complete before signaling shutdown to frontend
    ///
    /// The future returned will complete when all registered tasks have signaled completion via
    /// [`complete`][ShutdownToken::complete]
    pub async fn join(&self) {
        self.tasks.close();
        self.tasks.wait().await;
        self.frontend_shutdown.cancel();
    }

    /// Invoke application shutdown
    ///
    /// This will signal backend shutdown and wake the
    /// [`wait_for_shutdown`][ShutdownToken::wait_for_shutdown] future so that registered tasks can
    /// clean up before indicating completion.
    pub fn shutdown(&self) {
        self.backend_shutdown.cancel();
    }
}

/// A token that a component can obtain via [`register`][ShutdownManager::register]
///
/// This does not implement `Clone` because there should only be a single instance of a given
/// `ShutdownToken`. If you just need a copy of the `CancellationToken` for invoking shutdown, use
/// [`ShutdownToken::clone_cancellation_token`].
#[derive(Debug)]
pub struct ShutdownToken {
    token: CancellationToken,
    complete_tx: Mutex<Option<oneshot::Sender<()>>>,
}

impl ShutdownToken {
    /// Create a new [`ShutdownToken`]
    fn new(token: CancellationToken, complete_tx: oneshot::Sender<()>) -> Self {
        Self {
            token,
            complete_tx: Mutex::new(Some(complete_tx)),
        }
    }

    /// Trigger application shutdown due to some unrecoverable state
    pub fn trigger_shutdown(&self) {
        self.token.cancel();
    }

    /// Get a clone of the cancellation token for triggering shutdown
    pub fn clone_cancellation_token(&self) -> CancellationToken {
        self.token.clone()
    }

    /// Future that completes when [`ShutdownManager`] that issued this token is shutdown
    pub async fn wait_for_shutdown(&self) {
        self.token.cancelled().await;
    }

    /// Signal back to the [`ShutdownManager`] that the component that owns this token is finished
    /// cleaning up and it is safe for the process to exit
    ///
    /// # Implementation Note
    ///
    /// It is not required that registered components invoke this method, as the `ShutdownToken`
    /// type invokes this method on `Drop`.
    pub fn complete(&self) {
        if let Some(s) = self.complete_tx.lock().take() {
            let _ = s.send(());
        }
    }
}

/// `ShutdownToken` implements `Drop` such that the completion signal is guaranteed to be sent
/// to the `ShutdownManager`. This will prevent application hang on shutdown in the event that a
/// registered component fails before signaling completion.
impl Drop for ShutdownToken {
    fn drop(&mut self) {
        self.complete();
    }
}

#[cfg(test)]
mod tests {
    use std::{
        sync::atomic::{AtomicBool, Ordering},
        time::Duration,
    };

    use futures::FutureExt;
    use tokio_util::sync::CancellationToken;

    use crate::ShutdownManager;

    #[tokio::test]
    async fn test_shutdown_order() {
        let frontend_token = CancellationToken::new();
        let shutdown_manager = ShutdownManager::new(frontend_token.clone());

        static CLEAN: AtomicBool = AtomicBool::new(false);

        let token = shutdown_manager.register();
        tokio::spawn(async move {
            loop {
                futures::select! {
                    _ = token.wait_for_shutdown().fuse() => {
                        CLEAN.store(true, Ordering::SeqCst);
                        break;
                    }
                    _ = tokio::time::sleep(Duration::from_millis(10)).fuse() => {
                        // sleeping... 😴
                    }
                }
            }
        });

        shutdown_manager.shutdown();
        shutdown_manager.join().await;
        assert!(
            CLEAN.load(Ordering::SeqCst),
            "backend shutdown did not compelte"
        );
        assert!(
            frontend_token.is_cancelled(),
            "frontend shutdown was not triggered"
        );
    }
}
